#pragma once

#include <map>
#include <ctime>
#include <string>
#include <functional>

#include "nvs.h"



namespace SiliconTorch {

  namespace Metrics {

    typedef std::function<void()> Incrementer;

    uint32_t uptime_ms();  // rolls over after about 136 years

    class CounterMetric {
      public:
        CounterMetric(const std::string& shortName, const std::string& fullName);

        const std::string fullName;
        const std::string shortName;

        bool dirty = false;

        uint64_t getValue();

        void inc();
        void dec();

        void load(nvs_handle_t* nvs);       // load value from NVS
        void save(nvs_handle_t* nvs);       // saves only if enoughg time elapsed
        void forceSave(nvs_handle_t* nvs);  // saves value immediately

        uint64_t& operator++();    // pre-increment
        uint64_t& operator--();    // pre-decrement

        uint64_t operator++(int);  // post-increment
        uint64_t operator--(int);  // post-decrement

      private:

        uint64_t value = 0;

    };

    class Metrics {
      public:
        Metrics(const std::string& nameSpace);

        void registerMetric(const std::string& shortName, const std::string& fullName);

        /*
         * returns the actual value or 0 for unknown counters
         * also supports getting the special bootCnt, uptime and lifetime metrics
        */
        uint64_t getCounterValue(const std::string& shortName);

        Incrementer generateMetricIncrementer(const std::string& shortName);

        void saveData();
        void publishData();

        const std::string nameSpace;

        std::string genDeviceTopic(const char *suffix);

      private:

        uint32_t saveInterval      = 60*60;        // seconds
        uint32_t publishInterval   = 60;           // seconds

        uint32_t lastSaved         = uptime_ms();  // unix timestamp
        uint32_t lastPublished     = uptime_ms();  // unix timestamp

        uint32_t tickPeriod        = 3000;         // milli seconds

        uint64_t lifeTime          = 0;            // milli seconds
        uint64_t lifeTimeSaved     = 0;            // milli seconds

        uint32_t bootCnt           = 0;
        uint32_t bootCntSaveUptime = 60;           // seconds

        void tickTask();
        void saveBootCntTask();

        std::map<std::string, CounterMetric*> counters;

        nvs_handle_t nvs;

    };

  }
}
